@@ -18,7 +18,7 @@
 #include "popcount.h"
 
 /* locally defined functions */
-static void maybe_fire_collector PROTO((void));
+static void maybe_queue_fire_collector(thread_gc *tgc);
 
 void S_alloc_init() {
     ISPC s; IGEN g; UINT i;
@@ -33,9 +33,10 @@ void S_alloc_init() {
         for (g = 0; g <= static_generation; g++) {
             S_G.bytes_of_generation[g] = 0;
             for (s = 0; s <= max_real_space; s++) {
-              S_G.main_thread_gc.base_loc[g][s] = FIX(0); 
-              S_G.main_thread_gc.next_loc[g][s] = FIX(0); 
+              S_G.main_thread_gc.base_loc[g][s] = FIX(0);
+              S_G.main_thread_gc.next_loc[g][s] = FIX(0);
               S_G.main_thread_gc.bytes_left[g][s] = 0;
+              S_G.main_thread_gc.sweep_next[g][s] = NULL;
               S_G.bytes_of_space[g][s] = 0;
             }
         }
@@ -60,6 +61,10 @@ void S_alloc_init() {
         find_room(tc, space_new, 0, type_typed_object, size_fxvector(0), S_G.null_fxvector);
         FXVECTOR_TYPE(S_G.null_fxvector) = (0 << fxvector_length_offset) | type_fxvector;
 
+        S_protect(&S_G.null_flvector);
+        find_room(tc, space_new, 0, type_typed_object, size_flvector(0), S_G.null_flvector);
+        FXVECTOR_TYPE(S_G.null_flvector) = (0 << flvector_length_offset) | type_flvector;
+
         S_protect(&S_G.null_bytevector);
         find_room(tc, space_new, 0, type_typed_object, size_bytevector(0), S_G.null_bytevector);
         BYTEVECTOR_TYPE(S_G.null_bytevector) = (0 << bytevector_length_offset) | type_bytevector;
@@ -72,10 +77,6 @@ void S_alloc_init() {
         find_room(tc, space_new, 0, type_typed_object, size_vector(0), S_G.null_immutable_vector);
         VECTTYPE(S_G.null_immutable_vector) = (0 << vector_length_offset) | type_vector | vector_immutable_flag;
 
-        S_protect(&S_G.null_immutable_fxvector);
-        find_room(tc, space_new, 0, type_typed_object, size_fxvector(0), S_G.null_immutable_fxvector);
-        FXVECTOR_TYPE(S_G.null_immutable_fxvector) = (0 << fxvector_length_offset) | type_fxvector | fxvector_immutable_flag;
-
         S_protect(&S_G.null_immutable_bytevector);
         find_room(tc, space_new, 0, type_typed_object, size_bytevector(0), S_G.null_immutable_bytevector);
         BYTEVECTOR_TYPE(S_G.null_immutable_bytevector) = (0 << bytevector_length_offset) | type_bytevector | bytevector_immutable_flag;
@@ -96,7 +97,6 @@ void S_protect(p) ptr *p; {
     S_G.protected[S_G.protect_next++] = p;
 }
 
-/* S_reset_scheme_stack is always called with mutex */
 void S_reset_scheme_stack(tc, n) ptr tc; iptr n; {
     ptr *x; iptr m;
 
@@ -109,7 +109,7 @@ void S_reset_scheme_stack(tc, n) ptr tc; iptr n; {
         if (*x == snil) {
             if (n < default_stack_size) n = default_stack_size;
           /* stacks are untyped objects */
-            find_room(tc, space_new, 0, typemod, n, SCHEMESTACK(tc));
+            find_room(tc, space_new, 0, type_untyped, n, SCHEMESTACK(tc));
             break;
         }
         if ((m = CACHEDSTACKSIZE(*x)) >= n) {
@@ -137,6 +137,9 @@ ptr S_compute_bytes_allocated(xg, xs) ptr xg; ptr xs; {
   ISPC s, smax, smin; IGEN g, gmax, gmin;
   uptr n;
 
+  tc_mutex_acquire();
+  alloc_mutex_acquire();
+
   gmin = (IGEN)UNFIX(xg);
   if (gmin < 0) {
     gmin = 0;
@@ -180,6 +183,9 @@ ptr S_compute_bytes_allocated(xg, xs) ptr xg; ptr xs; {
   if (gmin == 0 && smin <= space_new && space_new <= smax)
       n -= (uptr)REAL_EAP(tc) - (uptr)AP(tc);
 
+  alloc_mutex_release();
+  tc_mutex_release();
+  
   return Sunsigned(n);
 }
 
@@ -187,12 +193,22 @@ ptr S_bytes_finalized() {
   return Sunsigned(S_G.bytes_finalized);
 }
 
-static void maybe_fire_collector() {
+/* called with alloc mutex */
+static void maybe_queue_fire_collector(thread_gc *tgc) {
   if ((S_G.bytes_of_generation[0] + S_G.bytesof[0][countof_phantom]) - S_G.g0_bytes_after_last_gc >= S_G.collect_trip_bytes)
-    S_fire_collector();
+    tgc->queued_fire = 1;
 }
 
-/* suitable mutex (either tc_mutex or gc_tc_mutex) must be held */
+void S_maybe_fire_collector(thread_gc *tgc) {
+  if ((tgc->during_alloc == 0) && (!IS_ALLOC_MUTEX_OWNER() || IS_TC_MUTEX_OWNER())) {
+    if (tgc->queued_fire) {
+      tgc->queued_fire = 0;
+      S_fire_collector();
+    }
+  }
+}
+
+/* allocation mutex must be held (or single-threaded guaranteed because collecting) */
 static void close_off_segment(thread_gc *tgc, ptr old, ptr base_loc, ptr sweep_loc, ISPC s, IGEN g)
 {
   if (base_loc) {
@@ -209,6 +225,9 @@ static void close_off_segment(thread_gc *tgc, ptr old, ptr base_loc, ptr sweep_l
     /* in case this is during a GC, add to sweep list */
     si = SegInfo(addr_get_segment(base_loc));
     si->sweep_start = sweep_loc;
+#if defined(WRITE_XOR_EXECUTE_CODE)
+    si->sweep_bytes = bytes;
+#endif
     si->sweep_next = tgc->sweep_next[g][s];
     tgc->sweep_next[g][s] = si;
   }
@@ -219,20 +238,13 @@ ptr S_find_more_gc_room(thread_gc *tgc, ISPC s, IGEN g, iptr n, ptr old) {
   ptr new;
   iptr new_bytes;
 
-#ifdef PTHREADS
-  if (S_use_gc_tc_mutex)
-    gc_tc_mutex_acquire();
-  else
-    tc_mutex_acquire();
-#else
-  tc_mutex_acquire();
-#endif
+  alloc_mutex_acquire();
 
   close_off_segment(tgc, old, tgc->base_loc[g][s], tgc->sweep_loc[g][s], s, g);
 
-  S_pants_down += 1;
+  tgc->during_alloc += 1;
 
-  nsegs = (uptr)(n + ptr_bytes + bytes_per_segment - 1) >> segment_offset_bits;
+  nsegs = (uptr)(n + allocation_segment_tail_padding + bytes_per_segment - 1) >> segment_offset_bits;
 
  /* block requests to minimize fragmentation and improve cache locality */
   if (s == space_code && nsegs < 16) nsegs = 16;
@@ -244,26 +256,29 @@ ptr S_find_more_gc_room(thread_gc *tgc, ISPC s, IGEN g, iptr n, ptr old) {
 
   tgc->base_loc[g][s] = new;
   tgc->sweep_loc[g][s] = new;
-  tgc->bytes_left[g][s] = (new_bytes - n) - ptr_bytes;
+  tgc->bytes_left[g][s] = (new_bytes - n) - allocation_segment_tail_padding;
   tgc->next_loc[g][s] = (ptr)((uptr)new + n);
 
-  if (g == 0 && S_pants_down == 1) maybe_fire_collector();
-
-  S_pants_down -= 1;
-
-#ifdef PTHREADS
-  if (S_use_gc_tc_mutex)
-    gc_tc_mutex_release();
-  else
-    tc_mutex_release();
-#else
-  tc_mutex_release();
+#if defined(WRITE_XOR_EXECUTE_CODE)
+  if (s == space_code) {
+    /* Ensure allocated code segments are writable. The caller should
+       already have bracketed the writes with calls to start and stop
+       so there is no need for a stop here. */
+    S_thread_start_code_write(tgc->tc, 0, 1, NULL, 0);
+  }
 #endif
-  
+
+  if (tgc->during_alloc == 1) maybe_queue_fire_collector(tgc);
+
+  tgc->during_alloc -= 1;
+
+  alloc_mutex_release();
+  S_maybe_fire_collector(tgc);
+
   return new;
 }
 
-/* tc_mutex must be held */
+/* allocation mutex must be held (or single-threaded guaranteed because collecting) */
 void S_close_off_thread_local_segment(ptr tc, ISPC s, IGEN g) {
   thread_gc *tgc = THREAD_GC(tc);
 
@@ -273,9 +288,11 @@ void S_close_off_thread_local_segment(ptr tc, ISPC s, IGEN g) {
   tgc->bytes_left[g][s] = 0;
   tgc->next_loc[g][s] = (ptr)0;
   tgc->sweep_loc[g][s] = (ptr)0;
+  tgc->sweep_next[g][s] = NULL;
 }
 
-/* S_reset_allocation_pointer is always called with mutex */
+/* S_reset_allocation_pointer is always called with allocation mutex
+   (or single-threaded guaranteed because collecting) */
 /* We always allocate exactly one segment for the allocation area, since
    we can get into hot water with formerly locked objects, specifically
    symbols and impure records, that cross segment boundaries.  This allows
@@ -287,10 +304,11 @@ void S_close_off_thread_local_segment(ptr tc, ISPC s, IGEN g) {
 
 void S_reset_allocation_pointer(tc) ptr tc; {
   iptr seg;
+  thread_gc *tgc = THREAD_GC(tc);
 
-  S_pants_down += 1;
+  tgc->during_alloc += 1;
 
-  seg = S_find_segments(THREAD_GC(tc), space_new, 0, 1);
+  seg = S_find_segments(tgc, space_new, 0, 1);
 
   /* NB: if allocate_segments didn't already ensure we don't use the last segment
      of memory, we'd have to reject it here so cp2-alloc can avoid a carry check for
@@ -303,19 +321,19 @@ void S_reset_allocation_pointer(tc) ptr tc; {
   S_G.bytes_of_space[0][space_new] += bytes_per_segment;
   S_G.bytes_of_generation[0] += bytes_per_segment;
 
-  if (S_pants_down == 1) maybe_fire_collector();
+  if (tgc->during_alloc == 1) maybe_queue_fire_collector(THREAD_GC(tc));
 
   AP(tc) = build_ptr(seg, 0);
   REAL_EAP(tc) = EAP(tc) = (ptr)((uptr)AP(tc) + bytes_per_segment);
 
-  S_pants_down -= 1;
+  tgc->during_alloc -= 1;
 }
 
 void S_record_new_dirty_card(thread_gc *tgc, ptr *ppp, IGEN to_g) {
   uptr card = (uptr)TO_PTR(ppp) >> card_offset_bits;
   dirtycardinfo *ndc;
 
-  gc_tc_mutex_acquire();
+  alloc_mutex_acquire();
   ndc = S_G.new_dirty_cards;
   if (ndc != NULL && ndc->card == card) {
     if (to_g < ndc->youngest) ndc->youngest = to_g;
@@ -327,9 +345,10 @@ void S_record_new_dirty_card(thread_gc *tgc, ptr *ppp, IGEN to_g) {
     ndc->next = next;
     S_G.new_dirty_cards = ndc;
   }
-  gc_tc_mutex_release();
+  alloc_mutex_release();
 }
 
+/* allocation mutex must be held (or only one thread due to call by collector) */
 FORCEINLINE void mark_segment_dirty(seginfo *si, IGEN from_g, IGEN to_g) {
   IGEN old_to_g = si->min_dirty_byte;
   if (to_g < old_to_g) {
@@ -355,7 +374,7 @@ void S_dirty_set(ptr *loc, ptr x) {
     seginfo *si = SegInfo(addr_get_segment(TO_PTR(loc)));
     if (si->use_marks) {
       /* GC must be in progress */
-      if (!IMMEDIATE(x)) {
+      if (!FIXMEDIATE(x)) {
         seginfo *t_si = SegInfo(ptr_get_segment(x));
         if (t_si->generation < si->generation)
           S_record_new_dirty_card(THREAD_GC(get_thread_context()), loc, t_si->generation);
@@ -363,13 +382,16 @@ void S_dirty_set(ptr *loc, ptr x) {
     } else {
       IGEN from_g = si->generation;
       if (from_g != 0) {
+        alloc_mutex_acquire();
         si->dirty_bytes[((uptr)TO_PTR(loc) >> card_offset_bits) & ((1 << segment_card_offset_bits) - 1)] = 0;
         mark_segment_dirty(si, from_g, 0);
+        alloc_mutex_release();
       }
     }
   }
 }
 
+/* only called by GC, so no other thread is running */
 void S_mark_card_dirty(uptr card, IGEN to_g) {
   uptr loc = card << card_offset_bits;
   uptr seg = addr_get_segment(loc);
@@ -381,7 +403,8 @@ void S_mark_card_dirty(uptr card, IGEN to_g) {
   }
 }
 
-/* scan remembered set from P to ENDP, transfering to dirty vector */
+/* scan remembered set from P to ENDP, transfering to dirty vector;
+   allocation mutex must be held */
 void S_scan_dirty(ptr *p, ptr *endp) {
   uptr this, last;
  
@@ -419,7 +442,7 @@ void S_scan_remembered_set() {
   ptr tc = get_thread_context();
   uptr ap, eap, real_eap;
 
-  tc_mutex_acquire();
+  alloc_mutex_acquire();
 
   ap = (uptr)AP(tc);
   eap = (uptr)EAP(tc);
@@ -438,7 +461,8 @@ void S_scan_remembered_set() {
     S_reset_allocation_pointer(tc);
   }
 
-  tc_mutex_release();
+  alloc_mutex_release();
+  S_maybe_fire_collector(THREAD_GC(tc));
 }
 
 /* S_get_more_room is called from genereated machine code when there is
@@ -453,7 +477,8 @@ void S_get_more_room() {
   ptr xp; uptr ap, type, size;
 
   xp = XP(tc);
-  if ((type = TYPEBITS(xp)) == 0) type = typemod;
+  type = TYPEBITS(xp);
+  if ((type_untyped != 0) && (type == 0)) type = type_untyped;
   ap = (uptr)UNTYPE(xp, type);
   size = (uptr)((iptr)AP(tc) - (iptr)ap);
 
@@ -466,14 +491,7 @@ ptr S_get_more_room_help(ptr tc, uptr ap, uptr type, uptr size) {
   eap = (uptr)EAP(tc);
   real_eap = (uptr)REAL_EAP(tc);
 
-#ifdef PTHREADS
-  if (S_use_gc_tc_mutex)
-    gc_tc_mutex_acquire();
-  else
-    tc_mutex_acquire();
-#else
-  tc_mutex_acquire();
-#endif
+  alloc_mutex_acquire();
 
   S_scan_dirty(TO_VOIDP(eap), TO_VOIDP(real_eap));
   eap = real_eap;
@@ -508,14 +526,8 @@ ptr S_get_more_room_help(ptr tc, uptr ap, uptr type, uptr size) {
     }
   }
 
-#ifdef PTHREADS
-  if (S_use_gc_tc_mutex)
-    gc_tc_mutex_release();
-  else
-    tc_mutex_release();
-#else
-  tc_mutex_release();
-#endif
+  alloc_mutex_release();
+  S_maybe_fire_collector(THREAD_GC(tc));
 
   return x;
 }
@@ -698,12 +710,28 @@ ptr S_fxvector(n) iptr n; {
     return p;
 }
 
-ptr S_bytevector(n) iptr n; {
-  return S_bytevector2(n, 0);
+ptr S_flvector(n) iptr n; {
+    ptr tc;
+    ptr p; iptr d;
+
+    if (n == 0) return S_G.null_flvector;
+
+    if ((uptr)n > (uptr)maximum_flvector_length)
+        S_error("", "invalid flvector size request");
+
+    tc = get_thread_context();
+
+    d = size_flvector(n);
+    newspace_find_room(tc, type_typed_object, d, p);
+    FLVECTOR_TYPE(p) = (n << flvector_length_offset) | type_flvector;
+    return p;
 }
 
-ptr S_bytevector2(n, immobile) iptr n; IBOOL immobile; {
-    ptr tc;
+ptr S_bytevector(n) iptr n; {
+  return S_bytevector2(get_thread_context(), n, space_new);
+}
+
+ptr S_bytevector2(tc, n, spc) ptr tc; iptr n; ISPC spc; {
     ptr p; iptr d;
 
     if (n == 0) return S_G.null_bytevector;
@@ -711,11 +739,9 @@ ptr S_bytevector2(n, immobile) iptr n; IBOOL immobile; {
     if ((uptr)n > (uptr)maximum_bytevector_length)
         S_error("", "invalid bytevector size request");
 
-    tc = get_thread_context();
-
     d = size_bytevector(n);
-    if (immobile)
-      find_room(tc, space_immobile_data, 0, type_typed_object, d, p);
+    if (spc != space_new)
+      find_room(tc, spc, 0, type_typed_object, d, p);
     else
       newspace_find_room(tc, type_typed_object, d, p);
     BYTEVECTOR_TYPE(p) = (n << bytevector_length_offset) | type_bytevector;
@@ -730,14 +756,6 @@ ptr S_null_immutable_vector() {
   return v;
 }
 
-ptr S_null_immutable_fxvector() {
-  ptr tc = get_thread_context();
-  ptr v;
-  find_room(tc, space_new, 0, type_typed_object, size_fxvector(0), v);
-  VECTTYPE(v) = (0 << fxvector_length_offset) | type_fxvector | fxvector_immutable_flag;
-  return v;
-}
-
 ptr S_null_immutable_bytevector() {
   ptr tc = get_thread_context();
   ptr v;
@@ -780,7 +798,7 @@ ptr Srecord_type(ptr r) {
 }
 
 ptr Srecord_type_parent(ptr rtd) {
-  return RECORDDESCPARENT(rtd);
+  return rtd_parent(rtd);
 }
 
 uptr Srecord_type_size(ptr rtd) {
@@ -1025,6 +1043,9 @@ ptr S_bignum(tc, n, sign) ptr tc; iptr n; IBOOL sign; {
     if ((uptr)n > (uptr)maximum_bignum_length)
         S_error("", "invalid bignum size request");
 
+    /* for anything that allocates bignums, make sure scheduling fuel is consumed */
+    USE_TRAP_FUEL(tc, n);
+
     d = size_bignum(n);
     newspace_find_room(tc, type_typed_object, d, p);
     BIGTYPE(p) = (uptr)n << bignum_length_offset | sign << bignum_sign_offset | type_bignum;
@@ -1050,7 +1071,7 @@ ptr S_relocation_table(n) iptr n; {
     ptr p; iptr d;
 
     d = size_reloc_table(n);
-    newspace_find_room(tc, typemod, d, p);
+    newspace_find_room(tc, type_untyped, d, p);
     RELOCSIZE(p) = n;
     return p;
 }
