@@ -21,7 +21,6 @@
 #include <sys/stat.h>
 #include <limits.h>
 #include <ctype.h>
-#include <time.h>
 
 /* locally defined functions */
 static INT s_errno PROTO((void));
@@ -38,6 +37,11 @@ static ptr s_ephemeron_pairp PROTO((ptr p));
 static ptr s_box_immobile PROTO((ptr p));
 static ptr s_make_immobile_vector PROTO((uptr len, ptr fill));
 static ptr s_make_immobile_bytevector PROTO((uptr len));
+static ptr s_make_reference_bytevector PROTO((uptr len));
+static ptr s_make_immobile_reference_bytevector PROTO((uptr len));
+static ptr s_reference_bytevectorp PROTO((ptr p));
+static ptr s_reference_star_address_object PROTO((ptr p));
+static ptr s_bytevector_reference_star_ref PROTO((ptr p, uptr offset));
 static ptr s_oblist PROTO((void));
 static ptr s_bigoddp PROTO((ptr n));
 static ptr s_float PROTO((ptr x));
@@ -53,7 +57,9 @@ static void s_showalloc PROTO((IBOOL show_dump, const char *outfn));
 static ptr s_system PROTO((const char *s));
 static ptr s_process PROTO((char *s, IBOOL stderrp));
 static I32 s_chdir PROTO((const char *inpath));
+#ifdef GETWD
 static char *s_getwd PROTO((void));
+#endif
 static ptr s_set_code_byte PROTO((ptr p, ptr n, ptr x));
 static ptr s_set_code_word PROTO((ptr p, ptr n, ptr x));
 static ptr s_set_code_long PROTO((ptr p, ptr n, ptr x));
@@ -97,6 +103,7 @@ static void s_mutex_acquire PROTO((scheme_mutex_t *m));
 static ptr s_mutex_acquire_noblock PROTO((scheme_mutex_t *m));
 static void s_condition_broadcast PROTO((s_thread_cond_t *c));
 static void s_condition_signal PROTO((s_thread_cond_t *c));
+static void s_thread_preserve_ownership PROTO((ptr tc));
 #endif
 static void s_byte_copy(ptr src, iptr srcoff, ptr dst, iptr dstoff, iptr cnt);
 static void s_ptr_copy(ptr src, iptr srcoff, ptr dst, iptr dstoff, iptr cnt);
@@ -208,7 +215,7 @@ static ptr s_box_immobile(p) ptr p; {
 }
 
 static ptr s_make_immobile_bytevector(uptr len) {
-  ptr b = S_bytevector2(len, 1);
+  ptr b = S_bytevector2(get_thread_context(), len, space_immobile_data);
   S_immobilize_object(b);
   return b;
 }
@@ -233,6 +240,36 @@ static ptr s_make_immobile_vector(uptr len, ptr fill) {
   return v;
 }
 
+static ptr s_make_reference_bytevector(uptr len) {
+  ptr b = S_bytevector2(get_thread_context(), len, space_reference_array);
+  memset(&BVIT(b, 0), 0, len);
+  return b;
+}
+
+static ptr s_make_immobile_reference_bytevector(uptr len) { 
+  ptr b = s_make_reference_bytevector(len);
+  S_immobilize_object(b);
+  return b;  
+}
+
+static ptr s_reference_bytevectorp(p) ptr p; {
+  seginfo *si;
+  return (si = MaybeSegInfo(ptr_get_segment(p))) != NULL && si->space == space_reference_array ? Strue : Sfalse;
+}
+
+static ptr s_reference_star_address_object(ptr p) {
+  if (p == (ptr)0)
+    return Sfalse;
+  else if (MaybeSegInfo(addr_get_segment(p)))
+    return (ptr)((uptr)p - reference_disp);
+  else
+    return Sunsigned((uptr)p);
+}
+
+static ptr s_bytevector_reference_star_ref(ptr p, uptr offset) {
+  return s_reference_star_address_object(*(ptr *)&BVIT(p, offset));
+}
+
 static ptr s_oblist() {
   ptr ls = Snil;
   iptr idx = S_G.oblist_length;
@@ -272,6 +309,10 @@ static ptr sorted_chunk_list(void) {
       ls = Scons(TO_PTR(chunk), ls);
       n += 1;
     }
+    for (chunk = (i == -1) ? S_code_chunks_full : S_code_chunks[i]; chunk != NULL; chunk = chunk->next) {
+      ls = Scons(TO_PTR(chunk), ls);
+      n += 1;
+    }
   }
 
   return sort_chunks(ls, n);
@@ -403,6 +444,7 @@ static void s_showalloc(IBOOL show_dump, const char *outfn) {
   ptr tc = get_thread_context();
 
   tc_mutex_acquire();
+  alloc_mutex_acquire();
 
   if (outfn == NULL) {
     out = stderr;
@@ -627,6 +669,7 @@ static void s_showalloc(IBOOL show_dump, const char *outfn) {
     fclose(out);
   }
 
+  alloc_mutex_release();
   tc_mutex_release();
 }
 
@@ -868,40 +911,59 @@ static char *s_getwd() {
 
 static ptr s_set_code_byte(p, n, x) ptr p, n, x; {
     I8 *a;
+    ptr tc = get_thread_context();
 
     a = (I8 *)TO_VOIDP((uptr)p + UNFIX(n));
+    S_thread_start_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I8));
     *a = (I8)UNFIX(x);
+    S_thread_end_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I8));
+
     return Svoid;
 }
 
 static ptr s_set_code_word(p, n, x) ptr p, n, x; {
     I16 *a;
+    ptr tc = get_thread_context();
 
     a = (I16 *)TO_VOIDP((uptr)p + UNFIX(n));
+    S_thread_start_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I16));
     *a = (I16)UNFIX(x);
+    S_thread_end_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I16));
+
     return Svoid;
 }
 
 static ptr s_set_code_long(p, n, x) ptr p, n, x; {
     I32 *a;
+    ptr tc = get_thread_context();
 
     a = (I32 *)TO_VOIDP((uptr)p + UNFIX(n));
+    S_thread_start_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I32));
     *a = (I32)(Sfixnump(x) ? UNFIX(x) : Sinteger_value(x));
+    S_thread_end_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I32));
+
     return Svoid;
 }
 
 static void s_set_code_long2(p, n, h, l) ptr p, n, h, l; {
     I32 *a;
+    ptr tc = get_thread_context();
 
     a = (I32 *)TO_VOIDP((uptr)p + UNFIX(n));
+    S_thread_start_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I32));
     *a = (I32)((UNFIX(h) << 16) + UNFIX(l));
+    S_thread_end_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I32));
 }
 
 static ptr s_set_code_quad(p, n, x) ptr p, n, x; {
     I64 *a;
+    ptr tc = get_thread_context();
 
     a = (I64 *)TO_VOIDP((uptr)p + UNFIX(n));
+    S_thread_start_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I64));
     *a = Sfixnump(x) ? UNFIX(x) : S_int64_value("\\#set-code-quad!", x);
+    S_thread_end_code_write(tc, 0, 0, TO_VOIDP(a), sizeof(I64));
+
     return Svoid;
 }
 
@@ -910,6 +972,7 @@ static ptr s_set_reloc(p, n, e) ptr p, n, e; {
 
     a = (iptr *)(&RELOCIT(CODERELOC(p), UNFIX(n)));
     *a = Sfixnump(e) ? UNFIX(e) : Sinteger_value(e);
+
     return e;
 }
 
@@ -921,8 +984,11 @@ static ptr s_flush_instruction_cache() {
 static ptr s_make_code(flags, free, name, arity_mark, n, info, pinfos)
                        iptr flags, free, n; ptr name, arity_mark, info, pinfos; {
     ptr co;
+    ptr tc = get_thread_context();
 
-    co = S_code(get_thread_context(), type_code | (flags << code_flags_offset), n);
+    S_thread_start_code_write(tc, 0, 0, NULL, 0);
+
+    co = S_code(tc, type_code | (flags << code_flags_offset), n);
     CODEFREE(co) = free;
     CODENAME(co) = name;
     CODEARITYMASK(co) = arity_mark;
@@ -931,12 +997,19 @@ static ptr s_make_code(flags, free, name, arity_mark, n, info, pinfos)
     if (pinfos != Snil) {
       S_G.profile_counters = Scons(S_weak_cons(co, pinfos), S_G.profile_counters);
     }
+
+    S_thread_end_code_write(tc, 0, 0, NULL, 0);
+
     return co;
 }
 
 static ptr s_make_reloc_table(codeobj, n) ptr codeobj, n; {
+    ptr tc = get_thread_context();
+
+    S_thread_start_code_write(tc, 0, 0, TO_VOIDP(&CODERELOC(codeobj)), sizeof(ptr));
     CODERELOC(codeobj) = S_relocation_table(UNFIX(n));
     RELOCCODE(CODERELOC(codeobj)) = codeobj;
+    S_thread_end_code_write(tc, 0, 0, TO_VOIDP(&CODERELOC(codeobj)), sizeof(ptr));
     return Svoid;
 }
 
@@ -1297,8 +1370,8 @@ static ptr s_set_collect_trip_bytes(n) ptr n; {
     return Svoid;
 }
 
-static void c_exit(UNUSED I32 status) {
-    S_abnormal_exit();
+static void c_exit(I32 status) {
+    exit(status);
 }
 
 #if defined(__STDC__) || defined(USE_ANSI_PROTOTYPES)
@@ -1315,6 +1388,9 @@ extern double log1p();
 #endif /* LOG1P */
 #endif /* defined(__STDC__) || defined(USE_ANSI_PROTOTYPES) */
 
+static double s_mod PROTO((double x, double y));
+static double s_mod(x, y) double x, y; { return fmod(x, y); }
+
 static double s_exp PROTO((double x));
 static double s_exp(x) double x; { return exp(x); }
 
@@ -1481,7 +1557,11 @@ static iptr s_backdoor_thread(p) ptr p; {
 }
 
 static ptr s_threads() {
-  return S_threads;
+  ptr ts;
+  tc_mutex_acquire();
+  ts = S_threads;
+  tc_mutex_release();
+  return ts;
 }
 
 static void s_mutex_acquire(m) scheme_mutex_t *m; {
@@ -1514,6 +1594,15 @@ static void s_condition_broadcast(s_thread_cond_t *c) {
 static void s_condition_signal(s_thread_cond_t *c) {
   s_thread_cond_signal(c);
 }
+
+/* called with tc mutex held */
+static void s_thread_preserve_ownership(ptr tc) {
+  if (!THREAD_GC(tc)->preserve_ownership) {
+    THREAD_GC(tc)->preserve_ownership = 1;
+    S_num_preserve_ownership_threads++;
+  }
+}
+
 #endif
 
 static ptr s_profile_counters(void) {
@@ -1593,6 +1682,7 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)condition_broadcast", (void *)s_condition_broadcast);
     Sforeign_symbol("(cs)condition_signal", (void *)s_condition_signal);
     Sforeign_symbol("(cs)condition_wait", (void *)S_condition_wait);
+    Sforeign_symbol("(cs)thread_preserve_ownership", (void *)s_thread_preserve_ownership);
 #endif
     Sforeign_symbol("(cs)s_addr_in_heap", (void *)s_addr_in_heap);
     Sforeign_symbol("(cs)s_ptr_in_heap", (void *)s_ptr_in_heap);
@@ -1605,6 +1695,11 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)box_immobile", (void *)s_box_immobile);
     Sforeign_symbol("(cs)make_immobile_vector", (void *)s_make_immobile_vector);
     Sforeign_symbol("(cs)make_immobile_bytevector", (void *)s_make_immobile_bytevector);
+    Sforeign_symbol("(cs)s_make_reference_bytevector", (void *)s_make_reference_bytevector);
+    Sforeign_symbol("(cs)s_make_immobile_reference_bytevector", (void *)s_make_immobile_reference_bytevector);
+    Sforeign_symbol("(cs)s_reference_bytevectorp", (void *)s_reference_bytevectorp);
+    Sforeign_symbol("(cs)s_reference_star_address_object", (void *)s_reference_star_address_object);
+    Sforeign_symbol("(cs)s_bytevector_reference_star_ref", (void *)s_bytevector_reference_star_ref);
     Sforeign_symbol("(cs)continuation_depth", (void *)S_continuation_depth);
     Sforeign_symbol("(cs)single_continuation", (void *)S_single_continuation);
     Sforeign_symbol("(cs)c_exit", (void *)c_exit);
@@ -1628,6 +1723,8 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)s_strings_to_gensym", (void *)s_strings_to_gensym);
     Sforeign_symbol("(cs)s_intern_gensym", (void *)S_intern_gensym);
     Sforeign_symbol("(cs)s_uninterned", (void *)S_uninterned);
+    Sforeign_symbol("(cs)symbol_hash32", (void *)S_symbol_hash32);
+    Sforeign_symbol("(cs)symbol_hash64", (void *)S_symbol_hash64);
     Sforeign_symbol("(cs)cputime", (void *)S_cputime);
     Sforeign_symbol("(cs)realtime", (void *)S_realtime);
     Sforeign_symbol("(cs)clock_gettime", (void *)S_clock_gettime);
@@ -1656,9 +1753,7 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)getpid", (void *)s_getpid);
     Sforeign_symbol("(cs)fasl_read", (void *)S_fasl_read);
     Sforeign_symbol("(cs)bv_fasl_read", (void *)S_bv_fasl_read);
-    Sforeign_symbol("(cs)to_vfasl", (void *)S_to_vfasl);
     Sforeign_symbol("(cs)vfasl_to", (void *)S_vfasl_to);
-    Sforeign_symbol("(cs)vfasl_can_combinep", (void *)S_vfasl_can_combinep);
     Sforeign_symbol("(cs)s_decode_float", (void *)s_decode_float);
 
     Sforeign_symbol("(cs)new_open_input_fd", (void *)S_new_open_input_fd);
@@ -1706,6 +1801,7 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)s_big_positive_bit_field", (void *)S_big_positive_bit_field);
     Sforeign_symbol("(cs)s_big_eq", (void *)S_big_eq);
     Sforeign_symbol("(cs)s_big_lt", (void *)S_big_lt);
+    Sforeign_symbol("(cs)s_big_trailing_zero_bits", (void *)S_big_trailing_zero_bits);
     Sforeign_symbol("(cs)s_bigoddp", (void *)s_bigoddp);
     Sforeign_symbol("(cs)s_div", (void *)S_div);
     Sforeign_symbol("(cs)s_float", (void *)s_float);
@@ -1723,6 +1819,7 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)s_set_random_seed", (void *)s_set_random_seed);
     Sforeign_symbol("(cs)ss_trunc", (void *)S_trunc);
     Sforeign_symbol("(cs)ss_trunc_rem", (void *)s_trunc_rem);
+    Sforeign_symbol("(cs)s_rational", (void *)S_rational);
     Sforeign_symbol("(cs)sub", (void *)S_sub);
     Sforeign_symbol("(cs)rem", (void *)S_rem);
 #ifdef GETWD
@@ -1737,6 +1834,7 @@ void S_prim5_init() {
     Sforeign_symbol("(cs)dequeue_scheme_signals", (void *)S_dequeue_scheme_signals);
     Sforeign_symbol("(cs)register_scheme_signal", (void *)S_register_scheme_signal);
 
+    Sforeign_symbol("(cs)mod", (void *)s_mod);
     Sforeign_symbol("(cs)exp", (void *)s_exp);
     Sforeign_symbol("(cs)log", (void *)s_log);
     Sforeign_symbol("(cs)pow", (void *)s_pow);
@@ -2028,7 +2126,12 @@ static void s_free(uptr addr) {
 }
 
 #ifdef FEATURE_ICONV
-#ifdef WIN32
+#ifdef DISABLE_ICONV
+# define iconv_t int
+#define ICONV_OPEN(to, from) -1
+#define ICONV(cd, in, inb, out, outb) -1
+#define ICONV_CLOSE(cd) -1
+#elif defined(WIN32)
 typedef void *iconv_t;
 typedef iconv_t (*iconv_open_ft)(const char *tocode, const char *fromcode);
 typedef size_t (*iconv_ft)(iconv_t cd, char **inbuf, size_t *inbytesleft, char **outbuf, size_t *outbytesleft);
