@@ -25,6 +25,8 @@ static void handle_call_error PROTO((ptr tc, iptr type, ptr x));
 static void init_signal_handlers PROTO((void));
 static void keyboard_interrupt PROTO((ptr tc));
 
+static void (*register_modified_signal)(int);
+
 ptr S_get_scheme_arg(tc, n) ptr tc; iptr n; {
 
     if (n <= asm_arg_reg_cnt) return REGARG(tc, n);
@@ -127,11 +129,8 @@ void S_split_and_resize() {
    * and clength + size(values) < stack-size; also, size may include
    * argument register values */
     n = CONTCLENGTH(k) + (value_count * sizeof(ptr)) + stack_slop;
-    if (n >= SCHEMESTACKSIZE(tc)) {
-       tc_mutex_acquire();
+    if (n >= SCHEMESTACKSIZE(tc))
        S_reset_scheme_stack(tc, n);
-       tc_mutex_release();
-    }
 }
 
 iptr S_continuation_depth(k) ptr k; {
@@ -272,7 +271,6 @@ void S_overflow(tc, frame_request) ptr tc; iptr frame_request; {
             }
 
           /* create a continuation */
-            tc_mutex_acquire();
             STACKLINK(tc) = S_mkcontinuation(space_new,
                                         0,
                                         CODEENTRYPOINT(nuate),
@@ -283,7 +281,6 @@ void S_overflow(tc, frame_request) ptr tc; iptr frame_request; {
                                         *split_point,
                                         Snil,
                                         Sfalse);
-            tc_mutex_release();
 
           /* overwrite old return address with dounderflow */
               *split_point = TO_PTR(DOUNDERFLOW);
@@ -296,9 +293,7 @@ void S_overflow(tc, frame_request) ptr tc; iptr frame_request; {
 
   /* allocate a new stack, retaining same relative sfp */
     sfp_offset = (uptr)TO_PTR(sfp) - (uptr)TO_PTR(split_point);
-    tc_mutex_acquire();
     S_reset_scheme_stack(tc, above_split_size + frame_request);
-    tc_mutex_release();
     SFP(tc) = (ptr)((uptr)SCHEMESTACK(tc) + sfp_offset);
 
   /* copy up everything above the split point.  we don't know where the
@@ -318,20 +313,21 @@ void S_error_abort(s) const char *s; {
 void S_abnormal_exit() {
   S_abnormal_exit_proc();
   fprintf(stderr, "abnormal_exit procedure did not exit\n");
-  exit(1);
+  abort();
 }
 
 static void reset_scheme() {
     ptr tc = get_thread_context();
 
-    tc_mutex_acquire();
+    alloc_mutex_acquire();
    /* eap should always be up-to-date now that we write-through to the tc
       when making any changes to eap when eap is a real register */
     S_scan_dirty(TO_VOIDP(EAP(tc)), TO_VOIDP(REAL_EAP(tc)));
     S_reset_allocation_pointer(tc);
     S_reset_scheme_stack(tc, stack_slop);
+    alloc_mutex_release();
     FRAME(tc,0) = TO_PTR(DOUNDERFLOW);
-    tc_mutex_release();
+    S_maybe_fire_collector(THREAD_GC(tc));
 }
 
 /* error_resets occur with the system in an unknown state,
@@ -391,12 +387,15 @@ static void do_error(type, who, s, args) iptr type; const char *who, *s; ptr arg
                        Scons(Sstring_utf8(s, -1), args)));
 
 #ifdef PTHREADS
-    while (S_tc_mutex_depth > 0) {
+    while (S_mutex_is_owner(&S_alloc_mutex))
+      S_mutex_release(&S_alloc_mutex);
+    while (S_mutex_is_owner(&S_tc_mutex))
       S_mutex_release(&S_tc_mutex);
-      S_tc_mutex_depth -= 1;
-    }
 #endif /* PTHREADS */
-    
+
+    /* in case error is during fasl read: */
+    S_thread_end_code_write(tc, static_generation, 0, NULL, 0);
+
     TRAP(tc) = (ptr)1;
     AC0(tc) = (ptr)1;
     CP(tc) = S_symbol_value(S_G.error_id);
@@ -511,7 +510,7 @@ void S_fire_collector() {
 
 /*  printf("firing collector!\n"); fflush(stdout); */
 
-  if (!Sboolean_value(S_symbol_value(crp_id))) {
+  if (!Sboolean_value(S_symbol_racy_value(crp_id))) {
     ptr ls;
 
 /*    printf("really firing collector!\n"); fflush(stdout); */
@@ -536,6 +535,10 @@ void S_noncontinuable_interrupt() {
   do_error(ERROR_NONCONTINUABLE_INTERRUPT,"","",Snil);
 }
 
+void Sscheme_register_signal_registerer(void (*registerer)(int)) {
+  register_modified_signal = registerer;
+}
+
 #ifdef WIN32
 ptr S_dequeue_scheme_signals(UNUSED ptr tc) {
   return Snil;
@@ -565,7 +568,7 @@ static BOOL WINAPI handle_signal(DWORD dwCtrlType) {
 #else
       ptr tc = get_thread_context();
 #endif
-      if (!S_pants_down && Sboolean_value(KEYBOARDINTERRUPTPENDING(tc)))
+      if (!THREAD_GC(tc)->during_alloc && Sboolean_value(KEYBOARDINTERRUPTPENDING(tc)))
         return(FALSE);
       keyboard_interrupt(tc);
       return(TRUE);
@@ -689,7 +692,7 @@ static void handle_signal(INT sig, UNUSED siginfo_t *si, UNUSED void *data) {
            /* disable keyboard interrupts in subordinate threads until we think
              of something more clever to do with them */
             if (tc == TO_PTR(&S_G.thread_context)) {
-              if (!S_pants_down && Sboolean_value(KEYBOARDINTERRUPTPENDING(tc))) {
+              if (!THREAD_GC(tc)->during_alloc && Sboolean_value(KEYBOARDINTERRUPTPENDING(tc))) {
                /* this is a no-no, but the only other options are to ignore
                   the signal or to kill the process */
                 RESET_SIGNAL
@@ -715,31 +718,42 @@ static void handle_signal(INT sig, UNUSED siginfo_t *si, UNUSED void *data) {
         case SIGBUS:
 #endif /* SIGBUS */
         case SIGSEGV:
+          {
+            ptr tc = get_thread_context();
             RESET_SIGNAL
-            if (S_pants_down)
+            if (THREAD_GC(tc)->during_alloc)
                 S_error_abort("nonrecoverable invalid memory reference");
             else
                 S_error_reset("invalid memory reference");
+          }
         default:
             RESET_SIGNAL
             S_error_reset("unexpected signal");
     }
 }
 
+static void no_op_register(UNUSED int sigid) {
+}
+
+#define SIGACTION(id, act_p, old_p) (register_modified_signal(id), sigaction(id, act_p, old_p))
+
 static void init_signal_handlers() {
     struct sigaction act;
 
+    if (register_modified_signal == NULL)
+      register_modified_signal = no_op_register;
+
     sigemptyset(&act.sa_mask);
 
   /* drop pending keyboard interrupts */
     act.sa_flags = 0;
     act.sa_handler = SIG_IGN;
-    sigaction(SIGINT, &act, (struct sigaction *)0);
+    SIGACTION(SIGINT, &act, (struct sigaction *)0);
 
   /* ignore broken pipe signals */
     act.sa_flags = 0;
     act.sa_handler = SIG_IGN;
-    sigaction(SIGPIPE, &act, (struct sigaction *)0);
+    SIGACTION(SIGPIPE, &act, (struct sigaction *)0);
 
   /* set up to catch SIGINT w/no system call restart */
 #ifdef SA_INTERRUPT
@@ -748,7 +762,7 @@ static void init_signal_handlers() {
     act.sa_flags = SA_SIGINFO;
 #endif /* SA_INTERRUPT */
     act.sa_sigaction = handle_signal;
-    sigaction(SIGINT, &act, (struct sigaction *)0);
+    SIGACTION(SIGINT, &act, (struct sigaction *)0);
 #ifdef BSDI
     siginterrupt(SIGINT, 1);
 #endif
@@ -760,14 +774,14 @@ static void init_signal_handlers() {
     act.sa_flags |= SA_RESTART;
 #endif /* SA_RESTART */
 #ifdef SIGQUIT
-    sigaction(SIGQUIT, &act, (struct sigaction *)0);
+    SIGACTION(SIGQUIT, &act, (struct sigaction *)0);
 #endif /* SIGQUIT */
-    sigaction(SIGILL, &act, (struct sigaction *)0);
-    sigaction(SIGFPE, &act, (struct sigaction *)0);
+    SIGACTION(SIGILL, &act, (struct sigaction *)0);
+    SIGACTION(SIGFPE, &act, (struct sigaction *)0);
 #ifdef SIGBUS
-    sigaction(SIGBUS, &act, (struct sigaction *)0);
+    SIGACTION(SIGBUS, &act, (struct sigaction *)0);
 #endif /* SIGBUS */
-    sigaction(SIGSEGV, &act, (struct sigaction *)0);
+    SIGACTION(SIGSEGV, &act, (struct sigaction *)0);
 }
 
 #endif /* WIN32 */
@@ -775,6 +789,7 @@ static void init_signal_handlers() {
 void S_schsig_init() {
     if (S_boot_time) {
         ptr p;
+        ptr tc = get_thread_context();
 
         S_protect(&S_G.nuate_id);
         S_G.nuate_id = S_intern((const unsigned char *)"$nuate");
@@ -786,13 +801,15 @@ void S_schsig_init() {
         S_protect(&S_G.collect_request_pending_id);
         S_G.collect_request_pending_id = S_intern((const unsigned char *)"$collect-request-pending");
 
-        p = S_code(get_thread_context(), type_code | (code_flag_continuation << code_flags_offset), 0);
+        S_thread_start_code_write(tc, 0, 0, NULL, 0);
+        p = S_code(tc, type_code | (code_flag_continuation << code_flags_offset), 0);
         CODERELOC(p) = S_relocation_table(0);
         CODENAME(p) = Sfalse;
         CODEARITYMASK(p) = FIX(0);
         CODEFREE(p) = 0;
         CODEINFO(p) = Sfalse;
         CODEPINFOS(p) = Snil;
+        S_thread_end_code_write(tc, 0, 0, NULL, 0);
 
         S_set_symbol_value(S_G.null_continuation_id,
             S_mkcontinuation(space_new,
@@ -820,7 +837,6 @@ void S_schsig_init() {
     }
 
 
-    S_pants_down = 0;
     S_set_symbol_value(S_G.collect_request_pending_id, Sfalse);
 
     init_signal_handlers();
