@@ -16,94 +16,33 @@
 
 #include "config.h"
 
-#if (machine_type == machine_type_arm32le || machine_type == machine_type_tarm32le || machine_type == machine_type_arm64le || machine_type == machine_type_tarm64le)
-# define OS_ANY_LINUX
-# if (machine_type == machine_type_tarm32le || machine_type == machine_type_tarm64le)
-#  define PTHREADS
-# endif
-# define OS_ANY_LINUX
-# define LITTLE_ENDIAN_IEEE_DOUBLE
+#if defined(scheme_feature_pthreads)
+# define PTHREADS
+#endif
+
+/*****************************************/
+/* Architectures                         */
+
+#if (defined(__powerpc__) || defined(__POWERPC__)) && !defined(__powerpc64__)
+# define PORTABLE_BYTECODE_BIGENDIAN
+# define BIG_ENDIAN_IEEE_DOUBLE
 # define FLUSHCACHE
 #endif
 
-#if (machine_type == machine_type_ppc32le || machine_type == machine_type_tppc32le || machine_type == machine_type_ppc64le || machine_type == machine_type_tppc64le)
-# define OS_ANY_LINUX
-# if (machine_type == machine_type_tppc32le || machine_type == machine_type_tppc64le)
-#  define PTHREADS
-# endif
+#if (defined(__arm__) || defined(__arm64__) || defined(__aarch64__))
 # define FLUSHCACHE
 #endif
 
-#if (machine_type == machine_type_i3le || machine_type == machine_type_ti3le || machine_type == machine_type_a6le || machine_type == machine_type_ta6le)
-# define OS_ANY_LINUX
-# if (machine_type == machine_type_ti3le || machine_type == machine_type_ta6le)
-#  define PTHREADS
-# endif
-# define LITTLE_ENDIAN_IEEE_DOUBLE
-#endif
-
-#if (machine_type == machine_type_i3fb || machine_type == machine_type_ti3fb || machine_type == machine_type_a6fb || machine_type == machine_type_ta6fb)
-# define OS_ANY_FREEBSD
-# if (machine_type == machine_type_ti3fb || machine_type == machine_type_ta6fb)
-#  define PTHREADS
-# endif
-# define LITTLE_ENDIAN_IEEE_DOUBLE
-#endif
-
-#if (machine_type == machine_type_i3nb || machine_type == machine_type_ti3nb || machine_type == machine_type_a6nb || machine_type == machine_type_ta6nb)
-# define OS_ANY_NETBSD
-# if (machine_type == machine_type_ti3nb || machine_type == machine_type_ta6nb)
-#  define PTHREADS
-# endif
-#endif
-
-#if (machine_type == machine_type_i3nt || machine_type == machine_type_ti3nt || machine_type == machine_type_a6nt || machine_type == machine_type_ta6nt)
-# define OS_ANY_WINDOWS
-# if (machine_type == machine_type_ti3nt || machine_type == machine_type_ta6nt)
-#  define PTHREADS
-# endif
-#endif
-
-#if (machine_type == machine_type_i3ob || machine_type == machine_type_ti3ob || machine_type == machine_type_a6ob || machine_type == machine_type_ta6ob)
-# define OS_ANY_OPENBSD
-# if (machine_type == machine_type_ti3ob || machine_type == machine_type_ta6ob)
-#  define PTHREADS
-# endif
-#endif
-
-#if (machine_type == machine_type_i3osx || machine_type == machine_type_ti3osx || machine_type == machine_type_a6osx || machine_type == machine_type_ta6osx)
-# define OS_ANY_MACOSX
-# if (machine_type == machine_type_ti3osx || machine_type == machine_type_ta6osx)
-#  define PTHREADS
-# endif
-#endif
-
 #if (machine_type == machine_type_pb)
-# if defined(__powerpc__) && !defined(__powerpc64__)
-#  define PORTABLE_BYTECODE_BIGENDIAN
-# endif
-# if defined(__linux__)
-#  define OS_ANY_LINUX
-#  ifndef PORTABLE_BYTECODE_BIGENDIAN
-#   define LITTLE_ENDIAN_IEEE_DOUBLE
-#  endif
-# elif defined(__NetBSD__)
-#  define OS_ANY_NETBSD
-# elif defined(__OpenBSD__) && !defined(__Bitrig__)
-#  define OS_ANY_OPENBSD
-# elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-#  define OS_ANY_FREEBSD
-#  ifndef PORTABLE_BYTECODE_BIGENDIAN
-#   define LITTLE_ENDIAN_IEEE_DOUBLE
-#  endif
-# elif defined(_MSC_VER) || defined(__MINGW32__)
-#   define OS_ANY_WINDOWS
-# elif __APPLE__
-#   define OS_ANY_MACOSX
-# endif
+# undef FLUSHCACHE
+#else
+# undef PORTABLE_BYTECODE_BIGENDIAN
 #endif
 
-#ifdef OS_ANY_LINUX
+/*****************************************/
+/* Operating systems                     */
+
+#if defined(__linux__)
 #define NOBLOCK O_NONBLOCK
 #define LOAD_SHARED_OBJECT
 #define USE_MMAP
@@ -116,7 +55,9 @@ typedef char *memcpy_t;
 #define MAKE_NAN(x) { x = 0.0; x = x / x; }
 #define GETWD(x) getcwd((x),PATH_MAX)
 typedef int tputsputcchar;
-#define LOCKF
+#ifndef __ANDROID__
+# define LOCKF
+#endif
 #define DIRMARKERP(c) ((c) == '/')
 #ifndef DISABLE_X11
 # define LIBX11 "libX11.so"
@@ -131,10 +72,14 @@ typedef int tputsputcchar;
 #define NSECCTIME(sb) (sb).st_ctim.tv_nsec
 #define NSECMTIME(sb) (sb).st_mtim.tv_nsec
 #define ICONV_INBUF_TYPE char **
+#ifdef __ANDROID__
+# define NOFILE 256
+# define NO_USELOCALE
+#endif
 #define UNUSED __attribute__((__unused__))
 #endif
 
-#ifdef OS_ANY_FREEBSD
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
 #define NOBLOCK O_NONBLOCK
 #define LOAD_SHARED_OBJECT
 #define USE_MMAP
@@ -163,7 +108,7 @@ typedef int tputsputcchar;
 #define USE_OSSP_UUID
 #endif
 
-#ifdef OS_ANY_NETBSD
+#if defined(__NetBSD__)
 #ifdef PTHREADS
 # define NETBSD
 #endif
@@ -172,7 +117,6 @@ typedef int tputsputcchar;
 #define USE_MMAP
 #define MMAP_HEAP
 #define IEEE_DOUBLE
-#define LITTLE_ENDIAN_IEEE_DOUBLE
 #define LDEXP
 #define ARCHYPERBOLIC
 #define GETPAGESIZE() getpagesize()
@@ -198,15 +142,15 @@ typedef int tputsputcchar;
 #define USE_MBRTOWC_L
 #endif
 
-#ifdef OS_ANY_WINDOWS
+#if defined(_MSC_VER) || defined(__MINGW32__)
 #define GETPAGESIZE() S_getpagesize()
 #define GETWD(x) GETCWD(x, _MAX_PATH)
 #define IEEE_DOUBLE
-#define LITTLE_ENDIAN_IEEE_DOUBLE
 #define LOAD_SHARED_OBJECT
 #define USE_VIRTUAL_ALLOC
 #define NAN_INCLUDE <math.h>
 #define MAKE_NAN(x) { x = sqrt(-1.0); }
+#define ARCHYPERBOLIC
 #ifndef PATH_MAX
 # define PATH_MAX _MAX_PATH
 #endif
@@ -261,17 +205,16 @@ struct timespec;
 #endif
 #if defined(__MINGW32__) && (machine_type == machine_type_ti3nt || machine_type == machine_type_i3nt)
 #define time_t __time64_t
-#define time _time64
+#define GET_TIME _time64
 #endif
 #endif
 
-#ifdef OS_ANY_OPENBSD
+#if defined(__OpenBSD__) && !defined(__Bitrig__)
 #define NOBLOCK O_NONBLOCK
 #define LOAD_SHARED_OBJECT
 #define USE_MMAP
 #define MMAP_HEAP
 #define IEEE_DOUBLE
-#define LITTLE_ENDIAN_IEEE_DOUBLE
 #define LDEXP
 #define ARCHYPERBOLIC
 #define GETPAGESIZE() getpagesize()
@@ -296,14 +239,29 @@ typedef int tputsputcchar;
 #define USE_OSSP_UUID
 #endif
 
-#ifdef OS_ANY_MACOSX
+#if defined(__APPLE__)
 #define MACOSX
 #define NOBLOCK O_NONBLOCK
 #define LOAD_SHARED_OBJECT
 #define USE_MMAP
 #define MMAP_HEAP
 #define IEEE_DOUBLE
-#define LITTLE_ENDIAN_IEEE_DOUBLE
+/* for both iPhone and iPhoneSimulator */
+#if defined(TARGET_OS_IPHONE)
+# define SYSTEM(s) ((void)s, -1)
+# define WRITE_XOR_EXECUTE_CODE
+#endif
+#if defined(__arm64__)
+# if !defined(WRITE_XOR_EXECUTE_CODE)
+#  define S_MAP_CODE MAP_JIT
+#  define S_ENABLE_CODE_WRITE(on) pthread_jit_write_protect_np(!(on))
+# endif
+# define CANNOT_READ_DIRECTLY_INTO_CODE
+# include <pthread.h>
+#elif defined(__x86_64__)
+/* needed to run under Rosetta2 on ARM Mac OS: */
+# define CANNOT_READ_DIRECTLY_INTO_CODE
+#endif
 #define LDEXP
 #define ARCHYPERBOLIC
 #define GETPAGESIZE() getpagesize()
@@ -314,7 +272,10 @@ typedef int tputsputcchar;
 #define LOCKF
 #define DIRMARKERP(c) ((c) == '/')
 #ifndef DISABLE_X11
-#define LIBX11 "/usr/X11R6/lib/libX11.dylib"
+# define DISABLE_X11
+#endif
+#ifndef DISABLE_X11
+# define LIBX11 "/usr/X11R6/lib/libX11.dylib"
 #endif
 #define _DARWIN_USE_64_BIT_INODE
 #define SECATIME(sb) (sb).st_atimespec.tv_sec
@@ -327,20 +288,12 @@ typedef int tputsputcchar;
 #define UNUSED __attribute__((__unused__))
 #endif
 
-#if (machine_type == machine_type_i3qnx || machine_type == machine_type_ti3qnx)
-# define OS_ANY_QNX
-# if (machine_type == machine_type_ti3qnx)
-#  define PTHREADS
-# endif
-#endif
-
-#ifdef OS_ANY_QNX
+#if defined(__QNX__)
 #define NOBLOCK O_NONBLOCK
 #define LOAD_SHARED_OBJECT
 #define USE_MMAP
 #define MMAP_HEAP
 #define IEEE_DOUBLE
-#define LITTLE_ENDIAN_IEEE_DOUBLE
 #define LDEXP
 #define ARCHYPERBOLIC
 #define GETPAGESIZE() getpagesize()
@@ -364,24 +317,17 @@ typedef int tputsputcchar;
 #define UNUSED
 #endif
 
-#if (machine_type == machine_type_i3s2 || machine_type == machine_type_ti3s2 || machine_type == machine_type_a6s2 || machine_type == machine_type_ta6s2)
-# define OS_ANY_SOLARIS2
-# if (machine_type == machine_type_ti3s2 || machine_type == machine_type_ta6s2)
-#  define PTHREADS
-# endif
-#endif
-
-#ifdef OS_ANY_SOLARIS2
+#if defined(sun)
 #define NOBLOCK O_NONBLOCK
 #define LOAD_SHARED_OBJECT
 #define USE_MMAP
 #define MMAP_HEAP
 #define IEEE_DOUBLE
-#define LITTLE_ENDIAN_IEEE_DOUBLE
 #define LDEXP
 #define ARCHYPERBOLIC
 #define LOG1P
 #define DEFINE_MATHERR
+#define NO_USELOCALE
 #define GETPAGESIZE() getpagesize()
 typedef char *memcpy_t;
 #define MAKE_NAN(x) { x = 0.0; x = x / x; }
@@ -403,7 +349,12 @@ typedef char tputsputcchar;
 #define UNUSED __attribute__((__unused__))
 #endif
 
-/* defaults */
+/*****************************************/
+/* Defaults and derived                  */
+
+#ifndef BIG_ENDIAN_IEEE_DOUBLE
+# define LITTLE_ENDIAN_IEEE_DOUBLE
+#endif
 
 #ifndef CHDIR
 # define CHDIR chdir
@@ -466,14 +417,45 @@ typedef char tputsputcchar;
 # define WRITE write
 #endif
 
+#ifndef S_PROT_CODE
+# ifdef WRITE_XOR_EXECUTE_CODE
+#  define S_PROT_CODE (PROT_WRITE | PROT_READ)
+# else
+#  define S_PROT_CODE (PROT_READ | PROT_WRITE | PROT_EXEC)
+# endif
+#endif
+#ifndef S_MAP_CODE
+# define S_MAP_CODE 0
+#endif
+#ifndef S_ENABLE_CODE_WRITE
+# define S_ENABLE_CODE_WRITE(on) do { } while (0)
+#endif
+
+/* WX_UNUSED indicates that an argument is used only for
+   WRITE_XOR_EXECUTE_CODE mode */
+#ifndef WX_UNUSED
+# ifdef WRITE_XOR_EXECUTE_CODE
+#  define WX_UNUSED
+# else
+#  define WX_UNUSED UNUSED
+# endif
+#endif
+
 #ifdef PTHREADS
 # define NO_THREADS_UNUSED /* empty */
 #else
 # define NO_THREADS_UNUSED UNUSED
 #endif
 
+#if defined(__has_feature)
+# if __has_feature(thread_sanitizer)
+#  define NO_THREAD_SANITIZE __attribute__((no_sanitize("thread")))
+#  define IMPLICIT_ATOMIC_AS_EXPLICIT
+# endif
+#endif
+#ifndef NO_THREAD_SANITIZE
+# define NO_THREAD_SANITIZE /* empty */
+#endif
+
 /* Use "/dev/urandom" everywhere except Windows */
 #define USE_DEV_URANDOM_UUID
-
-/* For debugging: */
-/* #define IMPLICIT_ATOMIC_AS_EXPLICIT */
