@@ -13,7 +13,7 @@ check_pb()
     DEST=$SCHEME_WORKAREA/pb/boot/pb/$1
     if [ ! -e $DEST ] ; then
         touch boot_pending
-    elif [ "`/usr/bin/find "$SRC" -newer "$DEST"`" != "" ] ; then
+    elif [ "`find "$SRC" -newer "$DEST"`" != "" ] ; then
         touch boot_pending
     fi
 }
@@ -25,7 +25,6 @@ check_pb equates.h
 check_pb gc-ocd.inc
 check_pb gc-oce.inc
 check_pb gc-par.inc
-check_pb vfasl.inc
 check_pb heapcheck.inc
 
 check_mach()
@@ -34,7 +33,7 @@ check_mach()
     INIT=$SCHEME_WORKAREA/boot/$MACH/$1
     if [ ! -e $INIT ] ; then
         touch boot_pending
-    elif [ "`/usr/bin/find "$SRC" -newer "$INIT"`" != "" ] ; then
+    elif [ "`find "$SRC" -newer "$INIT"`" != "" ] ; then
         touch boot_pending
     fi
 }
@@ -46,5 +45,4 @@ check_mach equates.h
 check_mach gc-ocd.inc
 check_mach gc-oce.inc
 check_mach gc-par.inc
-check_mach vfasl.inc
 check_mach heapcheck.inc
