@@ -20,7 +20,7 @@ static rktio_fd_t *finish_unix_fd_creation(rktio_t *rktio, int fd, int modes, rk
 #endif
 
 #ifdef RKTIO_USE_PENDING_OPEN
-static rktio_fd_t *open_via_thread(rktio_t *rktio, const char *filename, int modes, int flags);
+static rktio_fd_t *open_via_thread(rktio_t *rktio, const char *filename, int modes, int flags, int perm_bits);
 static int do_pending_open_release(rktio_t *rktio, struct open_in_thread_t *data, int close_fd);
 #endif
 
@@ -113,7 +113,7 @@ static rktio_fd_t *open_read(rktio_t *rktio, const char *filename, int modes)
 #endif
 }
 
-static rktio_fd_t *open_write(rktio_t *rktio, const char *filename, int modes)
+static rktio_fd_t *open_write(rktio_t *rktio, const char *filename, int modes, int perm_bits)
 {
 #ifdef RKTIO_SYSTEM_UNIX
   int fd;
@@ -130,19 +130,19 @@ static rktio_fd_t *open_write(rktio_t *rktio, const char *filename, int modes)
     flags |= O_EXCL;
 
   do {
-    fd = open(filename, flags | RKTIO_NONBLOCKING | RKTIO_BINARY, 0666);
+    fd = open(filename, flags | RKTIO_NONBLOCKING | RKTIO_BINARY, perm_bits);
   } while ((fd == -1) && (errno == EINTR));
 
   if (errno == ENXIO) {
     /* FIFO with no reader? */
 #ifdef RKTIO_USE_PENDING_OPEN
-    return open_via_thread(rktio, filename, modes, flags | RKTIO_BINARY);
+    return open_via_thread(rktio, filename, modes, flags | RKTIO_BINARY, perm_bits);
 #else
     /* Try opening in RW mode: */
     flags -= O_WRONLY;
     flags |= O_RDWR;
     do {
-      fd = open(filename, flags | RKTIO_NONBLOCKING | RKTIO_BINARY, 0666);
+      fd = open(filename, flags | RKTIO_NONBLOCKING | RKTIO_BINARY, perm_bits);
     } while ((fd == -1) && (errno == EINTR));
 #endif
   }
@@ -167,13 +167,16 @@ static rktio_fd_t *open_write(rktio_t *rktio, const char *filename, int modes)
 
   wp = WIDE_PATH_temp(filename);
   if (!wp) return NULL;
-
+    
   fd = CreateFileW(wp,
 		   GENERIC_WRITE | ((modes & RKTIO_OPEN_READ) ? GENERIC_READ : 0),
 		   FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
 		   NULL,
 		   hmode,
-		   FILE_FLAG_BACKUP_SEMANTICS, /* lets us detect directories in NT */
+		   (FILE_FLAG_BACKUP_SEMANTICS /* lets us detect directories in NT */
+                    | ((perm_bits & RKTIO_PERMISSION_WRITE)
+                       ? 0
+                       : FILE_ATTRIBUTE_READONLY)),
 		   NULL);
 
   if (fd == INVALID_HANDLE_VALUE) {
@@ -268,14 +271,19 @@ static rktio_fd_t *finish_unix_fd_creation(rktio_t *rktio, int fd, int modes, rk
 }
 #endif
 
-rktio_fd_t *rktio_open(rktio_t *rktio, const char *filename, int modes)
+rktio_fd_t *rktio_open_with_create_permissions(rktio_t *rktio, const char *filename, int modes, int perm_bits)
 {
   if (modes & RKTIO_OPEN_WRITE)
-    return open_write(rktio, filename, modes);
+    return open_write(rktio, filename, modes, perm_bits);
   else
     return open_read(rktio, filename, modes);
 }
 
+rktio_fd_t *rktio_open(rktio_t *rktio, const char *filename, int modes)
+{
+  return rktio_open_with_create_permissions(rktio, filename, modes, RKTIO_DEFAULT_PERM_BITS);
+}
+
 /*========================================================================*/
 /* File positions                                                         */
 /*========================================================================*/
@@ -435,6 +443,7 @@ typedef struct open_in_thread_t {
   pthread_cond_t ready_cond; /* wait until helper thread is ready (including cleanup) */
   char *filename;
   int flags;
+  int perm_bits;
   int done;
   int fd;
   int errval;
@@ -500,7 +509,7 @@ static void *do_open_in_thread(void *_data)
      canceled before `open` returns, but not both (otherwise there
      could be a space leak) */
   do {
-    fd = open(data->filename, data->flags, 0666);
+    fd = open(data->filename, data->flags, data->perm_bits);
   } while ((fd == -1) && (errno == EINTR));
 
   pthread_setcanceltype(PTHREAD_CANCEL_DISABLE, NULL);
@@ -519,7 +528,7 @@ static void *do_open_in_thread(void *_data)
   return NULL;
 }
 
-static rktio_fd_t *open_via_thread(rktio_t *rktio, const char *filename, int modes, int flags)
+static rktio_fd_t *open_via_thread(rktio_t *rktio, const char *filename, int modes, int flags, int perm_bits)
 {
   open_in_thread_t *data;
 
@@ -529,6 +538,7 @@ static rktio_fd_t *open_via_thread(rktio_t *rktio, const char *filename, int mod
 
   data->filename = strdup(filename);
   data->flags = flags;
+  data->perm_bits = perm_bits;
   pthread_mutex_init(&data->lock, NULL);
   pthread_cond_init(&data->ready_cond, NULL);
 
